package ex3;

import java.io.Closeable;
import java.io.IOException;
import java.net.InetSocketAddress;
import java.net.SocketAddress;
import java.nio.ByteBuffer;
import java.nio.channels.DatagramChannel;
import java.util.concurrent.atomic.AtomicBoolean;

public abstract class Process implements Runnable{

	private final int port;
	private volatile boolean running = true;
	
	private volatile DatagramChannel datagramChannel = null;
	
	private static final int MAX_PAYLOAD = 1024; 
	private ByteBuffer packet = ByteBuffer.allocate(MAX_PAYLOAD);
	
	
	
	
	public Process(int port) {
		super();
		this.port = port;
	}


	 /**
     * opens the UDP socket..
     * @throws IOException - may be error because port is in use..
     */
    private void open() throws IOException {
    	try {
	    	datagramChannel = DatagramChannel.open();
	    	datagramChannel.socket().bind(new InetSocketAddress(port));
	    	datagramChannel.configureBlocking(true);
		} catch (SecurityException se) {
			// Security exceptions can occur if for example we try to bind a port lower 
			// than 1024 without sufficient rights in a unixoid system
			throw new IOException(se);
		}
    }
    
    public void start() {
    	if (!running) { //TODO 2 Threads calling this start() concurrently could get both into this if-statement!  Fix it! Tip: AtomicBoolean/or synchromized block
    		running = true;
    		new Thread(this).start();
    	} 
    }
    
    public void stop() {
    	if (running) {
    		running = false;
    		close(datagramChannel);
    	}
    }

	public void run()  {
		while (running) {
			try {
				open();
				while (datagramChannel.isOpen()) {
					packet.clear();
					SocketAddress from = datagramChannel.receive(packet);
					if (from instanceof InetSocketAddress) {
						packet.flip();
						String message = new String(packet.array(),0,packet.remaining(),"UTF-8");
						receiveMessage(((InetSocketAddress)from).getPort(), message);
					} else {
						throw new IOException("Illegal address type received");
					}
				}
	
			} catch (IOException ioe) {
				ioe.printStackTrace();
			} finally {
				close(datagramChannel);
			}
		}
	}
	
	protected abstract void receiveMessage(int from,String message);
	
	protected void sendMessage(int to,String message) {
		try {
			datagramChannel.send(ByteBuffer.wrap(message.getBytes("UTF-8")), new InetSocketAddress("127.0.0.1",to));
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	
	private static void close(Closeable cl) {
		try {
			if (cl != null) {
				cl.close();
			}
		} catch (IOException e) {}
	}
}
